import os
import multiprocessing
import datetime
from meaningless.bible_web_extractor import WebExtractor
from meaningless.utilities import common
from meaningless.utilities.exceptions import UnsupportedTranslationError, InvalidPassageError


class BaseDownloader:
    """
    An downloader object that stores Bible passages into a local file
    """

    __translations_with_omitted_passages = {
        'ASV': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                'Luke 17:36', 'Luke 23:17',
                'John 5:4',
                'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                'Romans 16:24'],
        'EHV': ['Matthew 23:14',
                'Mark 15:28',
                'Luke 17:36',
                'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                'Romans 16:24'],
        'ESV': ['Matthew 12:47', 'Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                'Luke 17:36', 'Luke 23:17',
                'John 5:4',
                'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                'Romans 16:24'],
        'ESVUK': ['Matthew 12:47', 'Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                  'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                  'Luke 17:36', 'Luke 23:17',
                  'John 5:4',
                  'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                  'Romans 16:24'],
        'GW': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
               'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
               'Luke 17:36', 'Luke 23:17',
               'John 5:4',
               'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
               'Romans 16:24'],
        'ISV': ['Mark 15:28',
                'Luke 17:36',
                'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29'],
        'LEB': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                'Luke 17:36', 'Luke 23:17',
                'John 5:4',
                'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29'],
        'NRSV': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                 'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                 'Luke 17:36', 'Luke 23:17',
                 'John 5:4',
                 'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                 'Romans 16:24'],
        'NRSVUE': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                   'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                   'Luke 17:36', 'Luke 23:17',
                   'John 5:4',
                   'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                   'Romans 16:24'],
        'NLT': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                'Luke 17:36', 'Luke 23:17',
                'John 5:4',
                'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                'Romans 16:24'],
        'NASB': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                 'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 15:28',
                 'John 5:4',
                 'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                 'Romans 16:24'],
        'NET': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                'Mark 7:16', 'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                'Luke 17:36', 'Luke 23:17',
                'John 5:4',
                'Acts 8:37', 'Acts 15:34', 'Acts 28:29',
                'Romans 16:24'],
        # NIVUK has some omitted passages, but they are handled in the Web Extractor such that their contents can be
        # minimally deduced and thus not require the special passage handling logic in this class
        'NOG': ['Matthew 17:21', 'Matthew 18:11', 'Matthew 23:14',
                'Mark 9:44', 'Mark 9:46', 'Mark 11:26', 'Mark 15:28',
                'Luke 17:36', 'Luke 23:17',
                'John 5:4',
                'Acts 8:37', 'Acts 15:34', 'Acts 24:7', 'Acts 28:29',
                'Romans 16:24'],
    }

    def __init__(self, file_writing_function, translation='NIV', show_passage_numbers=True,
                 default_directory=os.getcwd(), strip_excess_whitespace=False, enable_multiprocessing=True,
                 use_ascii_punctuation=False, file_extension='', write_key_as_string=False):
        """
        :param file_writing_function: Function definition used to specify how to write to a given file.
                                      The function should only take 2 arguments, which are the file path to write to
                                      and the in-memory object being sourced (in that order).
        :type file_writing_function: callable[[str, dict], int]
        :param translation: Translation code for the particular passage. For example, 'NIV', 'ESV', 'NLT'
        :type translation: str
        :param show_passage_numbers: If True, any present passage numbers are preserved. Defaults to True.
        :type show_passage_numbers: bool
        :param default_directory: Directory containing the downloaded file.
                                  Defaults to the current working directory.
        :type default_directory: str
        :param strip_excess_whitespace: If True, passages don't retain leading & trailing whitespaces as well as
                                        newline characters. Defaults to False.
        :type strip_excess_whitespace: bool
        :param enable_multiprocessing: If True, downloads are performed using multiple daemon processes, resulting in
                                       lower download times by splitting computations among multiple CPU cores.
                                       Defaults to True.
        :type enable_multiprocessing: bool
        :param use_ascii_punctuation: When True, converts all Unicode punctuation characters into their ASCII
                                      counterparts. Defaults to False.
        :type use_ascii_punctuation: bool
        :param file_extension: File extension used when reading from a default file when file_path is not provided
        :type file_extension: str
        :param write_key_as_string: If True, specifies that all keys in the downloaded file are converted to strings.
               Defaults to False.
        :type write_key_as_string: bool
        """
        self.translation = translation
        self.show_passage_numbers = show_passage_numbers
        self.default_directory = default_directory
        self.strip_excess_whitespace = strip_excess_whitespace
        self.enable_multiprocessing = enable_multiprocessing
        self.use_ascii_punctuation = use_ascii_punctuation
        self.file_extension = file_extension
        self.file_writing_function = file_writing_function
        self.write_key_as_string = write_key_as_string

    def download_passage(self, book, chapter, passage, file_path=''):
        """
        Downloads a single passage as a file.

        The chapter and passage parameters will be automatically adjusted to the respective chapter and passage
        boundaries of the specified book, except in the case where passage is unreasonably high.

        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage: Passage number
        :type passage: int
        :param file_path: When specified, saves the file to this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: 1 if the download was successful. 0 if an error occurred.
        :rtype: int
        """
        return self.download_passage_range(book, chapter, passage, chapter, passage, file_path)

    def download_passages(self, book, chapter, passage_from, passage_to, file_path=''):
        """
        Downloads a range of passages of the same chapter as a file.

        Chapter and passage parameters will be automatically adjusted to the respective chapter and passage boundaries
        of the specified book, except in the case where passage_from is unreasonably high.

        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage_from: First passage number to get
        :type passage_from: int
        :param passage_to: Last passage number to get
        :type passage_to: int
        :param file_path: When specified, saves the file to this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: 1 if the download was successful. 0 if an error occurred.
        :rtype: int
        """
        return self.download_passage_range(book, chapter, passage_from, chapter, passage_to, file_path)

    def download_chapter(self, book, chapter, file_path=''):
        """
        Downloads a single chapter as a file.

        The chapter parameter will be automatically adjusted to the chapter boundaries of the specified book.

        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param file_path: When specified, saves the file to this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: 1 if the download was successful. 0 if an error occurred.
        :rtype: int
        """
        return self.download_passage_range(book, chapter, 1, chapter, common.get_end_of_chapter(), file_path)

    def download_chapters(self, book, chapter_from, chapter_to, file_path=''):
        """
        Downloads a range of passages from a specified chapter selection as a file.

        Chapter parameters will be automatically adjusted to the chapter boundaries of the specified book.

        :param book: Name of the book
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param file_path: When specified, saves the file to this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: 1 if the download was successful. 0 if an error occurred.
        :rtype: int
        """
        return self.download_passage_range(book, chapter_from, 1, chapter_to, common.get_end_of_chapter(), file_path)

    def download_book(self, book, file_path=''):
        """
        Downloads a specific book of the Bible and saves it as a file

        :param book: Name of the book
        :type book: str
        :param file_path: When specified, saves the file to this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: 1 if the download was successful. 0 if an error occurred.
        :rtype: int
        """
        return self.download_passage_range(book, 1, 1, common.get_chapter_count(book, self.translation),
                                           common.get_end_of_chapter(), file_path)

    def download_passage_range(self, book, chapter_from, passage_from, chapter_to, passage_to, file_path=''):
        """
        Downloads a range of passages from one specific passage to another passage as a file.

        Chapter and passage parameters will be automatically adjusted to the respective chapter and passage boundaries
        of the specified book, except in the case where passage_from is unreasonably high.

        :param book: Name of the book
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param passage_from: First passage number to get in the first chapter
        :type passage_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param passage_to: Last passage number to get in the last chapter
        :type passage_to: int
        :param file_path: When specified, saves the file to this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: 1 if the download was successful. 0 if an error occurred.
        :rtype: int
        """
        translation = self.translation.upper()
        if common.is_unsupported_translation(translation):
            raise UnsupportedTranslationError(translation)
        # Standardise letter casing with minimal impact to the resulting file
        book_name = book.title()

        if common.get_chapter_count(book_name, translation) <= 0:
            raise InvalidPassageError(book_name, chapter_from, passage_from, chapter_to, passage_to, translation)
        # Cap passage components to ensure input validity and minimise web requests by avoiding invalid chapters
        capped_chapter_from = common.get_capped_integer(chapter_from,
                                                        max_value=common.get_chapter_count(book_name, translation))
        capped_passage_from = common.get_capped_integer(passage_from)
        capped_chapter_to = common.get_capped_integer(chapter_to,
                                                      max_value=common.get_chapter_count(book_name, translation))
        capped_passage_to = common.get_capped_integer(passage_to)

        online_bible = WebExtractor(translation=translation, show_passage_numbers=self.show_passage_numbers,
                                    output_as_list=True, strip_excess_whitespace_from_list=self.strip_excess_whitespace,
                                    use_ascii_punctuation=self.use_ascii_punctuation)

        # Set up the base document with the root-level keys
        # Upon downloading a file, the top-level keys might be ordered differently to when they were inserted.
        # This is likely due to Python not sorting dictionary keys internally, but could be due to something else.
        # This does not affect the information contained in the downloaded file, but could affect file comparisons.
        document = {
            'Info': {
                'Language': common.get_translation_language(translation),
                'Translation': translation,
                'Timestamp': datetime.datetime.now().astimezone().isoformat(),
                'Meaningless': common.MEANINGLESS_VERSION
            },
            book_name: {}
        }

        # Don't initialise the thread pool unless the extractor has been set to use multiprocessing.
        # This logic could be already running in a daemon process, and initialising the pool will cause an error.
        process_pool = None
        if self.enable_multiprocessing:
            process_pool = multiprocessing.Pool()
        process_results = []

        # Range is extended by 1 to include chapter_to in the loop iteration
        chapter_range = range(capped_chapter_from, capped_chapter_to + 1)
        for chapter in chapter_range:
            passage_initial = 1
            passage_final = common.get_end_of_chapter()
            # Exclude a certain first half of the initial chapter based on where the passage start should be
            if chapter == capped_chapter_from:
                passage_initial = capped_passage_from
            # Exclude a certain last half of the last chapter based on where the passage end should be
            if chapter == capped_chapter_to:
                passage_final = capped_passage_to

            if self.enable_multiprocessing:
                # Asynchronously obtain each set of passages to reduce overall download time.
                # These are daemon processes, so these shouldn't block the program from exiting and should be
                # expected to be garbage collected if the main process is stopped.
                process = process_pool.apply_async(self._get_passages_dict, (online_bible, book_name, chapter,
                                                                             passage_initial, passage_final),
                                                   error_callback=self.__handle_exception_from_process)
                # Add the process result to the list and extract the value later to prioritise doing more work
                process_results.append(process)
            else:
                document[book_name][self.__key_cast(chapter)] = self._get_passages_dict(online_bible, book_name,
                                                                                        chapter, passage_initial,
                                                                                        passage_final)

        if self.enable_multiprocessing:
            # Close the pool manually, as the garbage collector might not dispose of this automatically
            process_pool.close()
            # Explicitly wait for the processes to finish up in case some processes have heavy workloads
            process_pool.join()
            # When multiprocessing, all process results should be retrieved as a batch operation to minimise
            # the total time cost associated with the "get" method for each result.
            document[book_name] = {self.__key_cast(chapter): process_results.pop(0).get() for chapter in chapter_range}

        if len(file_path) <= 0:
            file_location = os.path.join(self.default_directory, f'{book_name}{self.file_extension}')
        else:
            file_location = file_path
        return self.file_writing_function(file_location, document)

    @staticmethod
    def __handle_exception_from_process(exception):
        """
        A helper function that gets run an exception is received when downloading passages.
        This is not a top-level global function, as it is intended to be solely used within this class.

        :param exception: Exception object that is re-raised after this method completes.
        :type exception: object
        """
        # Clear the arguments to prevent "stacking" of exception error messages.
        # An empty tuple means the user has to search elsewhere in the stack trace to find the point of failure.
        exception.args = ()

    def _get_passages_dict(self, online_bible, book, chapter, passage_from, passage_to):
        """
        A helper function that obtains a range of passages and organises them as a dictionary for output.
        Not to be exposed as a usable method, as this function mostly exists so that passage retrieval can be done
        in a multi-processed way. Pre-pending the method name with double underscores causes referencing issues.

        :param online_bible: Instance of WebExtractor to use to download the passages
        :type online_bible: WebExtractor
        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage_from: First passage number to get
        :type passage_from: int
        :param passage_to: Last passage number to get
        :type passage_to: int
        :return: Dictionary of passages, keyed on passage number
        :rtype: dict
        """
        online_bible.output_as_list = True
        passage_list = online_bible.get_passages(book, chapter, passage_from, passage_to)
        # passage_num is the numerical representation of the Unicode passage number at the start of each passage
        passage_num = passage_from
        # self.translation is not used, since getting this far means that the translation used in the
        # earlier call to get_passages is guaranteed to be valid.
        is_translation_with_omitted_passages = online_bible.translation in \
            self.__translations_with_omitted_passages.keys()
        passages = {}
        for passage in passage_list:
            if is_translation_with_omitted_passages:
                # This logic handles translations that omit passages, and have are not considered as valid verse on
                # the Bible Gateway site. It works by checking the passage against the list of known omitted
                # passages for this particular translation, and assigning an empty string if it is omitted.
                # This is to ensure the passage key matches the actual passage contents,
                # regardless of translation.
                passage_string = f'{book} {chapter}:{passage_num}'
                if passage_string in self.__translations_with_omitted_passages[online_bible.translation]:
                    passages[self.__key_cast(passage_num)] = ''
                    # Since this passage isn't supposed to exist in the given translation but it is still registered
                    # in the file, the number is upped twice in this loop iteration - once for the omitted
                    # passage and once for the passage after the omitted passage (whose contents is accessible in
                    # this particular iteration)
                    passage_num += 1

            # First passage of the chapter may not always have a verse number.
            # Unclear if this is a formatting issue on the Bible Gateway site, but it is added for consistency.
            # This is not done on the web extractor due to the difficulty of selecting the first passage in an
            # arbitrary range.
            if passage_num == 1 and self.show_passage_numbers and not passage.startswith('¹ '):
                passage = f'¹ {passage}'

            passages[self.__key_cast(passage_num)] = passage
            passage_num += 1
        return passages

    def __key_cast(self, key):
        """
        A helper function to cast a dictionary key to a string or an integer.

        :param key: Dictionary key
        :type key: int or str
        :return: Type-casted dictionary key
        :rtype: int or str
        """
        return common.cast_to_str_or_int(key, self.write_key_as_string)



import os
import csv

# This is a collection of common methods used for interacting with CSV files.


def __get_header_list():
    """
    Returns the list of header fields prepended onto each CSV file.

    :return: The list of header fields.
    :rtype: str[]
    """
    return ['Book', 'Chapter', 'Passage', 'Text', 'Language', 'Translation', 'Timestamp', 'Meaningless']


def write(data_file, document):
    """
    A helper function to write to a CSV data file.
    Note that the input data must adhere to the following conventions:

    1. The input document is a dictionary.
    2. There is a top-level key called 'Info', with string values for the following keys:
       'Language', 'Translation', 'Timestamp', 'Meaningless'
    3. There is at least one other top-level key-value pair, mapping to a dictionary of dictionaries.
    4. There are no keys or values with the value set as None (excluding the sub-keys under the top-level 'Info' key).

    :param data_file: Path to the data file to write to
    :type data_file: str
    :param document: In-memory data structure, usually a dictionary
    :type document: dict
    :return: Returns 1 on success. Raises an exception when a write problem occurs.
    :rtype: int
    """
    # Only create the directory if it doesn't already exist. This is also to account for directories which are the
    # top level of a given drive (e.g. C:/) which can't be created by the file system due to denied access.
    data_directory = os.path.dirname(data_file)
    if not os.path.exists(data_directory):
        os.makedirs(data_directory, exist_ok=True)
    # Use UTF-8-BOM encoding to allow for Unicode characters to be written to the file, but also display Unicode
    # characters correctly when viewing the data in certain spreadsheet applications
    with open(data_file, 'w', newline='', encoding='utf-8-sig') as file:
        csv_writer = csv.writer(file, quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(__get_header_list())

        # Unlike other file interfaces, the metadata info is required to be provided by the input object.
        # This is to allow the CSV write function to determine the keys to access data for each row.
        if 'Info' not in document.keys():
            raise KeyError

        # Convert the mapping of info data into a list to make it easier to unpack the data on the CSV row
        info_fields = [document['Info'][info_field_key] for info_field_key in document['Info'].keys()]
        for book in document.keys():
            # The book name is not known, so obtain it by skipping all "non-book" keys
            if book in ['Info']:
                continue
            for chapter in document[book]:
                for passage in document[book][chapter]:
                    csv_writer.writerow([book, chapter, passage, document[book][chapter][passage], *info_fields])
    return 1


def read(data_file):
    """
    A helper function to read a CSV data file.
    Note that the file data must adhere to the following conventions:

    1. A header row is included as the first line.
    2. All data is contained within the first 8 columns.

    :param data_file: Path the data file to read
    :type data_file: str
    :return: Contents of the file as an object. Raises an exception when a read problem occurs.
    :rtype: dict
    """
    # Use UTF-8-BOM encoding to match the type used to write the file (assuming it was written with this file interface)
    output = {}
    with open(data_file, 'r', encoding='utf-8-sig') as file:
        csv_reader = csv.DictReader(file, fieldnames=__get_header_list())
        # Ignore the initial row of headers
        next(csv_reader)
        for row in csv_reader:
            book = row['Book']
            chapter = row['Chapter']
            passage = row['Passage']

            # Only assign the metadata once, since this should be the same on all rows anyway
            if 'Info' not in output:
                output['Info'] = {}
                for info_field_key in row.keys():
                    if info_field_key not in ['Book', 'Chapter', 'Passage', 'Text']:
                        output['Info'][info_field_key] = row[info_field_key]
            if book not in output:
                output[book] = {}
            if chapter not in output[book]:
                output[book][chapter] = {}
            if passage not in output[book][chapter]:
                output[book][chapter][passage] = row['Text']
    return output



from urllib.request import urlopen
from urllib.error import URLError
from time import sleep
import re

# This is a collection of helper methods used across the various modules.

MEANINGLESS_VERSION = '1.0.0'
'''
The current version of the Meaningless library.
'''


def is_unsupported_translation(translation):
    """
    A helper function to determine if the provided translation code is supported

    :param translation: Translation code
    :type translation: str
    :return: True = the translation is not supported, False = the translation is supported
    :rtype: bool

    >>> is_unsupported_translation('msg')
    True
    >>> is_unsupported_translation('NIV')
    False
    """
    # These translations are particularly difficult to extract information from due to them using
    # non-conventional page layouts compared to other translations: 'MOUNCE', 'VOICE', 'MSG', 'PHILLIPS'
    return translation.upper() not in BIBLE_TRANSLATIONS.keys()


def is_matching_translation(translation1, translation2):
    """
    A helper function to compare two translation strings and determine if they are equivalent.
    This will also consider any applicable translation aliases.

    :param translation1: First translation code
    :type translation1: str
    :param translation2: Second translation code
    :type translation2: str
    :return: True = both translations are equivalent, False = the translations are different
    :rtype: bool

    >>> is_matching_translation('NIV', 'NIV')
    True
    >>> is_matching_translation('NlT', 'nLt')
    True
    >>> is_matching_translation('NLT', 'NIV')
    False
    >>> is_matching_translation('NIV ', 'NIV')
    False
    >>> is_matching_translation('NRSV', 'NRSVUE')
    True

    """
    translation_first = translation1.upper()
    translation_second = translation2.upper()
    aliases_nrsv = ['NRSV', 'NRSVUE']
    return translation_first == translation_second or \
        (translation_first in aliases_nrsv and translation_second in aliases_nrsv)


def get_end_of_chapter():
    """
    A helper function to define the max. number of passages any chapter could possibly have.
    This is based on functionality present on Bible Gateway where a very large number is capped to the chapter's end.

    :return: A static number
    :rtype: int
    """
    return 9000


def get_chapter_count(book, translation='NIV'):
    """
    A helper function to return the number of chapters in a given book for a particular translation.

    :param book: Name of the book
    :type book: str
    :param translation: Translation code for the particular book. For example, 'NIV', 'ESV', 'NLT'
    :type translation: str
    :return: Number of chapters in the book. 0 usually means an invalid book or unsupported translation.
    :rtype: int

    >>> get_chapter_count('Ecclesiastes')
    12
    >>> get_chapter_count('Barnabas')
    0
    >>> get_chapter_count('Ecclesiastes', translation='msg')
    0
    >>> get_chapter_count('Juan', 'RVA')
    21
    """
    bible_translation = translation.upper()
    bible_book = book.title()

    if bible_book == 'Song Of Solomon':
        # Song Of Songs has an alternate name
        bible_book = 'Song Of Songs'
    elif bible_book == 'Psalms':
        # Psalm and its plural variation are basically the same book, but prefer the singular variant
        bible_book = 'Psalm'
    elif bible_book == 'Phillippians':
        # Most, if not all, translations use the spelling with one L but also accept the alternative spelling
        bible_book = 'Philippians'

    if bible_translation in BIBLE_TRANSLATIONS.keys() and \
       bible_book in BIBLE_TRANSLATIONS[bible_translation]['Books'].keys():
        return BIBLE_TRANSLATIONS[bible_translation]['Books'][bible_book]
    return 0


def get_page(url, retry_count=3, retry_delay=2):
    """
    A helper function that returns the contents of a web page.

    :param url: Page URL to obtain
    :type url: str
    :param retry_count: Number of attempts to resend the request if it fails after the first try
    :type retry_count: int
    :param retry_delay: Number of seconds to wait before retrying a request. This increases after every retry.
    :type retry_delay: int
    :return: Page contents. Raises an error if the web page could not be loaded for any reason.
    :rtype: str

    >>> get_page('https://www.biblegateway.com')
    b'<!DOCTYPE html>...'
    >>> get_page('https://www.randomwebsite.com', retry_count=0)
    Traceback (most recent call last):
    ...
    urllib.error.URLError: <urlopen error ...>
    >>> get_page('https://pypi.org/project/meaningless/meaningless/', retry_count=2, retry_delay=1)
    Traceback (most recent call last):
    ...
    urllib.error.HTTPError: HTTP Error 404: Not Found
    """
    # Cap the values to ensure the function isn't suspended for an eternity, but still attempts at least once
    retries = get_capped_integer(retry_count, 0, 10)
    delay = get_capped_integer(retry_delay, 0, 30)
    delay_multiplier = 2
    # The extra addition to the range end is to account for the initial request
    for retry in range(0, retries + 1):
        try:
            with urlopen(url) as response:
                return response.read()
        except URLError as exception:
            if retry < retries:
                sleep(delay)
                delay *= delay_multiplier
                continue
            raise exception


def superscript_numbers(text, remove_brackets=True):
    """
    A helper function that converts a string's numeric characters into their superscript Unicode variations

    :param text: String to process
    :type text: str
    :param remove_brackets: If True, removes bracket characters that would be considered non-standard formatting.
                            Mostly used to handle the case of empty passages such as Luke 17:36. Defaults to True.
    :type remove_brackets: bool
    :rtype: str

    >>> superscript_numbers('[0123456789]')
    '⁰¹²³⁴⁵⁶⁷⁸⁹'
    >>> superscript_numbers('Antidisestablishmentarianism')
    'Antidisestablishmentarianism'
    >>> superscript_numbers('[7]', False)
    '[⁷]'
    """
    superscript_text = text
    if remove_brackets:
        # The strip method can't be relied on here, as the string itself can be space padded at times.
        # Using sequential replacements should be OK, unless there are more characters to handle.
        superscript_text = superscript_text.replace('[', '').replace(']', '').replace('(', '').replace(')', '')
    superscript_translation_table = superscript_text.maketrans('0123456789', '⁰¹²³⁴⁵⁶⁷⁸⁹')
    return superscript_text.translate(superscript_translation_table)


def remove_superscript_numbers_in_passage(text):
    """
    A helper function that removes all superscript numbers with optional trailing space from a given string.
    Mainly used to hide passage numbers in a given block of text.

    :param text: String to process
    :type text: str
    :return: String with the superscript numbers that have a trailing space removed
    :rtype: str

    >>> remove_superscript_numbers_in_passage('⁰ ¹ ² ³ ⁴ ⁵ ⁶ ⁷ ⁸ ⁹ ')
    ''
    >>> remove_superscript_numbers_in_passage('E=mc²')
    'E=mc'
    """
    return re.sub(r'[⁰¹²³⁴⁵⁶⁷⁸⁹]+\s?', '', text)


def get_capped_integer(number, min_value=1, max_value=100):
    """
    A helper function to limit an integer between an upper and lower bound

    :param number: Number to keep limited
    :type number: int or str
    :param min_value: Lowest possible value assigned when number is lower than this
    :type min_value: int or str
    :param max_value: Highest possible value assigned when number is larger than this
    :type max_value: int or str
    :return: Integer that adheres to min_value <= number <= max_value
    :rtype: int

    >>> get_capped_integer(42)
    42
    >>> get_capped_integer(0, min_value=7)
    7
    >>> get_capped_integer(42, max_value=7)
    7
    >>> get_capped_integer('0', min_value='7')
    7
    >>> get_capped_integer('42', max_value='7')
    7
    """
    return min(max(int(number), int(min_value)), int(max_value))


def dict_keys_to_sorted_list(keys):
    """
    A helper function that converts a dictionary's keys to a sorted list.

    :param keys: Dictionary keys
    :type keys: object
    :return: Dictionary keys casted as a sorted list
    :rtype: list

    >>> dict_keys_to_sorted_list({4: '', 2: '', 7: ''}.keys())
    [2, 4, 7]
    >>> dict_keys_to_sorted_list({'4': '', '2': '', '7': ''}.keys())
    ['2', '4', '7']
    """
    def key_sorting_function(key):
        # An inner sorting function that is primarily used for sorting numeric keys or strings of numeric characters.
        # TODO Sorting on string keys with actual alphabetic characters requires additional logic and doctests
        return int(key)

    sorted_keys = list(keys)
    sorted_keys.sort(key=key_sorting_function)
    return sorted_keys


def get_translation_language(translation):
    """
    A helper function to provide the language used in a given Bible translation

    :param translation: Translation code for the particular passage. For example, 'NIV', 'ESV', 'NLT'
    :type translation: str
    :return: Language used as a string. Empty string if the translation is not supported.
    :rtype: str

    >>> get_translation_language('NIV')
    'English'
    >>> get_translation_language('RVA')
    'Español'
    >>> get_translation_language('mounce')
    ''
    """
    bible_translation = translation.upper()
    if bible_translation in BIBLE_TRANSLATIONS.keys():
        return BIBLE_TRANSLATIONS[bible_translation]['Language']
    return ''


def unicode_to_ascii_punctuation(text):
    """
    A helper function to convert certain punctuation characters from Unicode to ASCII

    :param text: String to process
    :type text: str
    :return: String with ASCII punctuation where relevant
    :rtype: str

    >>> unicode_to_ascii_punctuation('‘GG’')
    "'GG'"
    >>> unicode_to_ascii_punctuation('“G—G”')
    '"G-G"'
    >>> unicode_to_ascii_punctuation('GG')
    'GG'
    >>> unicode_to_ascii_punctuation('⁰')
    '⁰'
    """
    punctuation_map = text.maketrans('“‘—’”', '"\'-\'"')
    return text.translate(punctuation_map)


def cast_to_str_or_int(text, cast_to_str):
    """
    A helper function to convert a given input to either a string or an integer.

    :param text: Input text
    :type text: int or str
    :param cast_to_str: When True, input text is cast to a string. When False, input is cast to an integer.
    :type cast_to_str: bool
    :return: A string or integer representation of the input string
    :rtype: int (str if cast_to_str is True)

    >>> cast_to_str_or_int('Tri-Beam', True)
    'Tri-Beam'
    >>> cast_to_str_or_int('42', False)
    42
    >>> cast_to_str_or_int('42', True)
    '42'
    >>> cast_to_str_or_int(42, False)
    42
    >>> cast_to_str_or_int(42, True)
    '42'
    >>> cast_to_str_or_int('Tri-Beam', False)
    'Tri-Beam'
    """
    if not cast_to_str:
        # Input text that cannot be cast to an integer are instead cast to a string
        try:
            return int(text)
        except ValueError:
            pass
    return str(text)


def get_bible_data_for_language(language, mode=0):
    """
    A helper function to return miscellaneous Bible data for a particular supported language

    :param language: Language
    :type language: str
    :param mode: Numeric value corresponding to a specific data set. Defaults to 0.
                 0 = All information
                 1 = New Testament only
    :type mode: int
    :return: Dictionary containing Bible data relating to a particular supported language
    :rtype: dict

    >>> get_bible_data_for_language('English')
    {'Language': 'English', 'Books': {...'Ruth': 4...}
    >>> get_bible_data_for_language('English', mode=1)
    {'Language': 'English', 'Books': {'Matthew': 28...}
    >>> get_bible_data_for_language('Español')
    {'Language': 'Español', 'Books': {...'Rut': 4...}
    >>> get_bible_data_for_language('Saiyan')
    {'Language': 'Saiyan', 'Books': {}}
    """
    bible_language = language.title()

    # This is a general mapping of books and their total chapters applicable to most translations of a specific
    # language. Apocrypha books are not listed, partly because their total chapters vary between translations but
    # mostly due to these books not being officially supported.
    bible_book_mapping = {
        'English': {
            'OT': {
                'Genesis': 50,
                'Exodus': 40,
                'Leviticus': 27,
                'Numbers': 36,
                'Deuteronomy': 34,
                'Joshua': 24,
                'Judges': 21,
                'Ruth': 4,
                '1 Samuel': 31,
                '2 Samuel': 24,
                '1 Kings': 22,
                '2 Kings': 25,
                '1 Chronicles': 29,
                '2 Chronicles': 36,
                'Ezra': 10,
                'Nehemiah': 13,
                'Esther': 10,
                'Job': 42,
                'Psalm': 150,
                'Proverbs': 31,
                'Ecclesiastes': 12,
                'Song Of Songs': 8,
                'Isaiah': 66,
                'Jeremiah': 52,
                'Lamentations': 5,
                'Ezekiel': 48,
                'Daniel': 12,
                'Hosea': 14,
                'Joel': 3,
                'Amos': 9,
                'Obadiah': 1,
                'Jonah': 4,
                'Micah': 7,
                'Nahum': 3,
                'Habakkuk': 3,
                'Zephaniah': 3,
                'Haggai': 2,
                'Zechariah': 14,
                'Malachi': 4
            },
            'NT': {
                'Matthew': 28,
                'Mark': 16,
                'Luke': 24,
                'John': 21,
                'Acts': 28,
                'Romans': 16,
                '1 Corinthians': 16,
                '2 Corinthians': 13,
                'Galatians': 6,
                'Ephesians': 6,
                'Philippians': 4,
                'Colossians': 4,
                '1 Thessalonians': 5,
                '2 Thessalonians': 3,
                '1 Timothy': 6,
                '2 Timothy': 4,
                'Titus': 3,
                'Philemon': 1,
                'Hebrews': 13,
                'James': 5,
                '1 Peter': 5,
                '2 Peter': 3,
                '1 John': 5,
                '2 John': 1,
                '3 John': 1,
                'Jude': 1,
                'Revelation': 22
            }
        },
        'Español': {
            'OT': {
                'Génesis': 50,
                'Éxodo': 40,
                'Levítico': 27,
                'Números': 36,
                'Deuteronomio': 34,
                'Josué': 24,
                'Jueces': 21,
                'Rut': 4,
                '1 Samuel': 31,
                '2 Samuel': 24,
                '1 Reyes': 22,
                '2 Reyes': 25,
                '1 Crónicas': 29,
                '2 Crónicas': 36,
                'Esdras': 10,
                'Nehemías': 13,
                'Ester': 10,
                'Job': 42,
                'Salmos': 150,
                'Proverbios': 31,
                'Eclesiastés': 12,
                'Cantares': 8,
                'Isaías': 66,
                'Jeremías': 52,
                'Lamentaciones': 5,
                'Ezequiel': 48,
                'Daniel': 12,
                'Oseas': 14,
                'Joel': 3,
                'Amós': 9,
                'Abdías': 1,
                'Jonás': 4,
                'Miqueas': 7,
                'Nahúm': 3,
                'Habacuc': 3,
                'Sofonías': 3,
                'Hageo': 2,
                'Zacarías': 14,
                'Malaquías': 4
            },
            'NT': {
                'Mateo': 28,
                'Marcos': 16,
                'Lucas': 24,
                'Juan': 21,
                'Hechos': 28,
                'Romanos': 16,
                '1 Corintios': 16,
                '2 Corintios': 13,
                'Gálatas': 6,
                'Efesios': 6,
                'Filipenses': 4,
                'Colosenses': 4,
                '1 Tesalonicenses': 5,
                '2 Tesalonicenses': 3,
                '1 Timoteo': 6,
                '2 Timoteo': 4,
                'Tito': 3,
                'Filemón': 1,
                'Hebreos': 13,
                'Santiago': 5,
                '1 Pedro': 5,
                '2 Pedro': 3,
                '1 Juan': 5,
                '2 Juan': 1,
                '3 Juan': 1,
                'Judas': 1,
                'Apocalipsis': 22
            }
        }
    }

    if bible_language in bible_book_mapping.keys():
        if mode == 1:
            bible_books = bible_book_mapping[bible_language]['NT']
        else:
            # Default to all information when the mode doesn't match one of the preset data sets
            bible_books = {**bible_book_mapping[bible_language]['OT'], **bible_book_mapping[bible_language]['NT']}
    else:
        # Unsupported language
        bible_books = {}

    return {
         'Language': bible_language,
         'Books': bible_books
    }


BIBLE_TRANSLATIONS = {
    # English
    'ASV': get_bible_data_for_language('English'),
    'AKJV': get_bible_data_for_language('English'),
    'BRG': get_bible_data_for_language('English'),
    'EHV': get_bible_data_for_language('English'),
    'ESV': get_bible_data_for_language('English'),
    'ESVUK': get_bible_data_for_language('English'),
    'GNV': get_bible_data_for_language('English'),
    'GW': get_bible_data_for_language('English'),
    'ISV': get_bible_data_for_language('English'),
    'JUB': get_bible_data_for_language('English'),
    'KJV': get_bible_data_for_language('English'),
    'KJ21': get_bible_data_for_language('English'),
    'LEB': get_bible_data_for_language('English'),
    'MEV': get_bible_data_for_language('English'),
    'NASB': get_bible_data_for_language('English'),
    'NASB1995': get_bible_data_for_language('English'),
    'NET': get_bible_data_for_language('English'),
    'NIV': get_bible_data_for_language('English'),
    'NIVUK': get_bible_data_for_language('English'),
    'NKJV': get_bible_data_for_language('English'),
    'NLT': get_bible_data_for_language('English'),
    'NLV': get_bible_data_for_language('English'),
    'NMB': get_bible_data_for_language('English', mode=1),
    'NOG': get_bible_data_for_language('English'),
    'NRSV': get_bible_data_for_language('English'),  # Contains Apocrypha books that are currently omitted
    'NRSVUE': get_bible_data_for_language('English'),  # Contains Apocrypha books that are currently omitted
    'WEB': get_bible_data_for_language('English'),
    'YLT': get_bible_data_for_language('English'),
    # Spanish
    'RVA': get_bible_data_for_language('Español')
}
'''
The mapping of supported translations and associated Bible data, excluding Apocrypha information. For simplicity, all
book names use their common variant (e.g. 'Song Of Songs' instead of 'Song Of Solomon').
'''

if __name__ == "__main__":
    # Run this section when run as a standalone script. Don't run this part when being imported.
    import doctest
    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)



import os
from meaningless.bible_base_extractor import BaseExtractor
from meaningless.utilities import csv_file_interface


class CSVExtractor(BaseExtractor):
    """
    An base extractor object that retrieves Bible passages from a CSV file
    """
    def __init__(self, translation='NIV', show_passage_numbers=True, output_as_list=False,
                 strip_excess_whitespace_from_list=False, default_directory=os.getcwd(),
                 use_ascii_punctuation=False):
        super().__init__(csv_file_interface.read, translation, show_passage_numbers, output_as_list,
                         strip_excess_whitespace_from_list, default_directory, use_ascii_punctuation,
                         file_extension='.csv', read_key_as_string=True)



# This file is used as a way of defining accessible modules when importing this library.
# Even though some modules are omitted, they can still be imported using their full name.
# e.g. meaningless.utilities.common

# Ignore the base extractor and downloader, as the only reason to expose it is to support custom extractors and
# downloaders, which is currently not supported or advised.
from meaningless.bible_yaml_downloader import YAMLDownloader
from meaningless.bible_yaml_extractor import YAMLExtractor
from meaningless.bible_json_downloader import JSONDownloader
from meaningless.bible_json_extractor import JSONExtractor
from meaningless.bible_xml_downloader import XMLDownloader
from meaningless.bible_xml_extractor import XMLExtractor
from meaningless.bible_csv_downloader import CSVDownloader
from meaningless.bible_csv_extractor import CSVExtractor
from meaningless.bible_web_extractor import WebExtractor
# Ignore the base error class, but include all the other exception types
from meaningless.utilities.exceptions import (
    UnsupportedTranslationError,
    InvalidPassageError,
    InvalidSearchError,
    TranslationMismatchError
)
# Include the file interfaces, mainly as an out-of-the-box mechanism for reading downloaded files
# as well as writing output using the information obtained from the extractors.
from meaningless.utilities import yaml_file_interface, json_file_interface, xml_file_interface, csv_file_interface



from bs4 import BeautifulSoup
from urllib.parse import urlencode
import re
from meaningless.utilities import common
from meaningless.utilities.exceptions import InvalidSearchError, UnsupportedTranslationError


class WebExtractor:
    """
    An extractor object that retrieves Bible passages from the Bible Gateway site.

    This does NOT extend from the BaseExtractor class, as it would expose certain attributes and function parameters
    that don't make sense for the Web Extractor and should not be interacted with (e.g. default directory, file path,
    file extension, etc.).
    """

    def __init__(self, translation='NIV', show_passage_numbers=True, output_as_list=False,
                 strip_excess_whitespace_from_list=False, use_ascii_punctuation=False):
        """
        :param translation: Translation code for the particular passage. For example, 'NIV', 'ESV', 'NLT'
        :type translation: str
        :param show_passage_numbers: If True, any present passage numbers are preserved. Defaults to True.
        :type show_passage_numbers: bool
        :param output_as_list: When True, returns the passage data as a list of strings. Defaults to False.
        :type output_as_list: bool
        :param strip_excess_whitespace_from_list: When True and output_as_list is also True, leading and trailing
                                                  whitespace characters are removed for each string element in the list.
                                                  Defaults to False.
        :type strip_excess_whitespace_from_list: bool
        :param use_ascii_punctuation: When True, converts all Unicode punctuation characters into their ASCII
                                      counterparts. This also applies to passage separators. Defaults to False.
        :type use_ascii_punctuation: bool
        """
        self.translation = translation
        self.show_passage_numbers = show_passage_numbers
        self.output_as_list = output_as_list
        self.strip_excess_whitespace_from_list = strip_excess_whitespace_from_list
        self.use_ascii_punctuation = use_ascii_punctuation

    def get_passage(self, book, chapter, passage):
        """
        Gets a single passage from the Bible Gateway site.

        The chapter and passage parameters will be automatically adjusted to the respective chapter and passage
        boundaries of the specified book.

        :param book: Name of the book (This must match the name used by the translation)
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage: Passage number
        :type passage: int
        :return: The specified passage. Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter, passage, chapter, passage)

    def get_passages(self, book, chapter, passage_from, passage_to):
        """
        Gets a range of passages of the same chapter from the Bible Gateway site.

        Chapter and passage parameters will be automatically adjusted to the respective chapter and passage boundaries
        of the specified book.

        :param book: Name of the book (This must match the name used by the translation)
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage_from: First passage number to get
        :type passage_from: int
        :param passage_to: Last passage number to get
        :type passage_to: int
        :return: The passages between the specified passages (inclusive). Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter, passage_from, chapter, passage_to)

    def get_chapter(self, book, chapter):
        """
        Gets a single chapter from the Bible Gateway site.

        The chapter parameter will be automatically adjusted to the chapter boundaries of the specified book.

        :param book: Name of the book (This must match the name used by the translation)
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :return: All passages in the chapter. Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter, 1, chapter, common.get_end_of_chapter())

    def get_chapters(self, book, chapter_from, chapter_to):
        """
        Gets a range of passages from a specified chapters selection from the Bible Gateway site.

        The chapter parameters will be automatically adjusted to the chapter boundaries of the specified book.

        :param book: Name of the book (This must match the name used by the translation)
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :return: All passages between the specified chapters (inclusive). Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter_from, 1, chapter_to, common.get_end_of_chapter())

    def get_book(self, book):
        """
        Gets all chapters for a specific book from the Bible Gateway site.

        :param book: Name of the book (This must match the name used by the translation)
        :type book: str
        :return: All passages in the specified book. Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, 1, 1, common.get_chapter_count(book, self.translation),
                                      common.get_end_of_chapter())

    def get_passage_range(self, book, chapter_from, passage_from, chapter_to, passage_to):
        """
        Gets a range of passages from one specific passage to another passage from the Bible Gateway site.

        Chapter and passage parameters will be automatically adjusted to the respective chapter and passage boundaries
        of the specified book.

        :param book: Name of the book (This must match the name used by the translation)
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param passage_from: First passage number to get in the first chapter
        :type passage_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param passage_to: Last passage number to get in the last chapter
        :type passage_to: int
        :return: All passages between the specified passages (inclusive). Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        # Capping the chapter and passage information, as this gets included in site search string and can cause
        # the web request to stagger if this manages to be long enough.
        capped_chapter_from = common.get_capped_integer(chapter_from,
                                                        max_value=common.get_chapter_count(book, self.translation))
        capped_passage_from = common.get_capped_integer(passage_from)
        capped_chapter_to = common.get_capped_integer(chapter_to,
                                                      max_value=common.get_chapter_count(book, self.translation))
        capped_passage_to = common.get_capped_integer(passage_to)
        # Defer to a direct search invocation when sourcing passages from the same chapter
        if capped_chapter_from == capped_chapter_to:
            return self.search(f'{book} {capped_chapter_from}:{capped_passage_from} - {capped_passage_to}')

        # Get the partial section of the first chapter being requested, omitting some initial passages
        initial_chapter = self.get_passages(book, capped_chapter_from, capped_passage_from, common.get_end_of_chapter())
        # Get the partial section of the last chapter being requested, omitting some trailing passages
        final_chapter = self.get_passages(book, capped_chapter_to, 1, capped_passage_to)
        # Get all the chapters in between the initial and final chapters (exclusive since they have been pre-fetched).
        # Sandwich those chapters between the first and last pre-fetched chapters to combine all the passage data.
        chapters = [initial_chapter] + \
                   [self.get_chapter(book, chapter)
                    for chapter in range(capped_chapter_from + 1, capped_chapter_to)] + [final_chapter]
        if self.output_as_list:
            # Flattens the data structure from a list of lists to a normal list
            return [chapter for chapter_list in chapters for chapter in chapter_list]
        return '\n'.join(chapters)

    def search_multiple(self, passage_names):
        """
        Retrieves a set of passages directly from the Bible Gateway site. Passages can be from different books.
        The language used for the search text is independent of the translation.
        Note that the output is subject to the Bible Gateway implicit passage limit when sending the web request.

        :param passage_names: List of Bible passages that are valid when used on www.biblegateway.com
        :type passage_names: list
        :return: Bible passages with newline separators for each set of passages
        :rtype: str or list
        """
        return self.search(';'.join(passage_names))

    def search(self, passage_name):
        """
        Retrieves a specific passage directly from the Bible Gateway site.
        The language used for the search text is independent of the translation.
        Note that the output is subject to the Bible Gateway implicit passage limit when sending the web request.

        :param passage_name: Name of the Bible passage which is valid when used on www.biblegateway.com
        :type passage_name: str
        :return: Bible passage with preserved line breaks
        :rtype: str or list
        """
        # Some translations are very tricky to extract passages from, and currently, so specific extraction logic
        # for these translations should not be introduced until they need to be supported.
        translation = self.translation.upper()
        if common.is_unsupported_translation(translation):
            raise UnsupportedTranslationError(translation)

        # Use the printer-friendly view since there are fewer page elements to load and process
        source_site_params = urlencode({'version': self.translation, 'search': passage_name, 'interface': 'print'})
        source_site = f'https://www.biblegateway.com/passage/?{source_site_params}'
        soup = BeautifulSoup(common.get_page(source_site), 'html.parser')

        # Don't collect contents from an invalid verse, since they do not exist.
        # A fail-fast approach can be taken by checking for certain indicators of invalidity.
        if not soup.find('div', {'class': 'passage-content'}):
            raise InvalidSearchError(source_site)

        # To get a list, the passage separator is given an actual practical use as an indicator of where to split
        # the string to create list elements.
        if self.output_as_list:
            passage_separator = '-_-'
        else:
            passage_separator = ''

        # Compile the list of tags to remove from the parsed web page, corresponding to the following elements:
        # h1
        #    - Ignore passage display
        # h2
        #    - Ignore chapter headings
        # h3
        #    - Ignore section headings
        # h4
        #    - Ignore subsection headings, such as those in Ezekiel 40
        # a with 'full-chap-link' class
        #    - Ignore the "Read Full Chapter" text, which is carefully embedded within the passage
        # sup with 'crossreference' class
        #    - Ignore cross references
        # sup with 'footnote' class
        #    - Ignore in-line footnotes
        # div with one of the 'footnotes', 'dropdowns', 'crossrefs', 'passage-other-trans' classes
        #    - Ignore the footer area, which is composed of several main tags
        # p with 'translation-note' class
        #    - Ignore explicit translation notes in translations such as ESV
        # crossref
        #    - Ignore in-line references in translations such as WEB
        removable_tags = soup.find_all(re.compile('^h1$|^h2$|^h3$|^h4$')) \
            + soup.find_all('a', {'class': re.compile('^full-chap-link$|^bibleref$')}) \
            + soup.find_all('sup', {'class': re.compile('^crossreference$|^footnote$')}) \
            + soup.find_all('div', {
                            'class': re.compile('^footnotes$|^dropdowns$|^crossrefs$|^passage-other-trans$')}) \
            + soup.find_all('p', {'class': re.compile('^translation-note$')}) \
            + soup.find_all('crossref')
        # Normally, paragraphs with the 'first-line-none' class would contain valid passage contents.
        # In the GNV translation, this class name is specifically used for the blurb of notable chapter details.
        if translation == 'GNV':
            removable_tags += soup.find_all('p', {'class': re.compile('^first-line-none$')})
        [tag.decompose() for tag in removable_tags]

        # Compile a list of ways Psalm interludes can be found. These are to be preserved, as it would be the
        # translation team's decision to omit these from the passage, as opposed to a code-level design decision.
        # span with 'selah' class
        #    - Explicit Psalm interludes in the translations such as NLT and CEB
        # i with 'selah' class
        #    - Explicit Psalm interludes in the translations such as NKJV
        # selah
        #    - Explicit Psalm interludes in the translations such as HCSB
        interludes = soup.find_all('span', {'class': 'selah'}) \
            + soup.find_all('i', {'class': 'selah'}) \
            + soup.find_all('selah')
        # Psalm interludes may or may not have a leading space, depending on the translation.
        # In any case, always add one in so that the interlude doesn't meld into the passage contents.
        # To account for double spaces, this relies on a regex replacement later on to convert down to a single space.
        [interlude.replace_with(f' {interlude.text}') for interlude in interludes]
        # <br> tags will naturally be ignored when getting text
        [br.replace_with('\n') for br in soup.find_all('br')]
        # The versenum tag appears in only a few translations such as NIVUK, and is difficult to handle because
        # its child tags are usually decomposed before this point, but space padding seems to take its place.
        # Interestingly, the tag itself has attributes that indicate what the actual passage number is, so this is
        # extracted instead of the actual text content contained within the tag.
        if translation == 'NIVUK':
            [versenum.replace_with(f'{passage_separator}{common.superscript_numbers(versenum.attrs["id"])}')
                for versenum in soup.find_all('versenum')]
        # Convert chapter numbers into new lines
        [chapter_num.replace_with('\n') for chapter_num in soup.find_all('span', {'class': 'chapternum'})]
        # Preserve superscript verse numbers by using their Unicode counterparts
        # Add in the custom passage separator as well while access to the verse numbers is still available
        [sup.replace_with(f'{passage_separator}{common.superscript_numbers(sup.text)}')
         for sup in soup.find_all('sup', {'class': 'versenum'})]
        # Some verses such as Nehemiah 7:30 - 42 store text in a <table> instead of <p>, which means
        # spacing is not preserved when collecting the text. Therefore, a space is manually injected
        # onto the end of the left cell's text to stop it from joining the right cell's text.
        # Note: Python "double colon" syntax for lists is used to retrieve items at every N interval including 0.
        # TODO: If a verse with >2 columns is found, this WILL need to be updated to be more dynamic
        [td.replace_with(f'{td.text} ') for td in soup.find_all('td')[::2]]
        # Preserve paragraph spacing by manually pre-pending a new line
        # THIS MUST BE THE LAST PROCESSING STEP because doing this earlier interferes with other replacements
        [p.replace_with(f'\n{p.text}') for p in soup.find_all('p')]

        # Combine the text contents of all passage sections on the page.
        # Convert non-breaking spaces to normal spaces when retrieving the raw passage contents.
        # Also strip excess whitespaces to prevent a whitespace build-up when combining multiple passages.
        #
        # Double square brackets are removed here, as they are mostly just indicators that the passages is only kept
        # due to convention with earlier translations. This replacement is done here, as it can sometimes have a
        # trailing space which can cause double spacing, which needs to be normalised.
        raw_passage_text = '\n'.join([tag.text.replace('\xa0', ' ').strip() for tag in
                                      soup.find_all('div', {'class': 'passage-content'})]) \
            .replace('[[', '').replace(']]', '')
        # To account for spaces between tags that end up blending into the passage contents, this regex replacement is
        # specifically used to remove that additional spacing, since it is part of the actual page layout.
        all_text = re.sub('([^ ]) {2,3}([^ ])', r'\1 \2', raw_passage_text)

        # Remove all superscript numbers if the passage numbers should be hidden
        if not self.show_passage_numbers:
            all_text = common.remove_superscript_numbers_in_passage(all_text)
        # Since the passage separator is prepended on all passages, the first occurrence is to be removed
        # as the separator is only needed between passages.

        # Perform ASCII punctuation conversion after hiding superscript numbers to process a slightly shorter string
        if self.use_ascii_punctuation:
            all_text = common.unicode_to_ascii_punctuation(all_text)

        # Some translations include asterisks on certain words in the New Testament. This usually indicates an in-line
        # marker that the word has been translated from present-tense Greek to past-tense English for better flow
        # in modern usage, though not all translations provide consistent footnotes on what the asterisk implies.
        # As it is not actually part of the passage text itself, this is expected to be ignored.
        # Also note that this is a naive replacement - fortunately, asterisks do not seem to be used as a proper
        # text character anywhere in the currently supported Bible translations.
        all_text = all_text.replace('*', '')
        # Translations such as GW add these text markers around certain words, which can be removed
        #
        # Translations such as JUB append a pilcrow character at the start of certain passages, which can be removed.
        # These usually have a trailing space, which also needs to be removed to prevent double spacing.
        # This logic would need to be revisited if there are cases of pilcrows without a trailing space.
        all_text = all_text.replace('⌞', '').replace('⌟', '').replace('¶ ', '')

        if not self.output_as_list:
            # Do any final touch-ups to the passage contents before outputting the string
            return all_text.strip()

        # At this point, the expectation is that the return value is a list of passages.
        # Since the passage separator is placed before the passage number, it can cause an empty first item upon
        # splitting if the first passage has a passage number.
        # Remove the first passage separator in such scenarios to prevent this from happening.
        # Also perform a one-off strip to ensure the first passage separator is correctly identified.
        if all_text.strip().startswith(passage_separator):
            all_text = all_text.replace(passage_separator, '', 1)
        passage_list = re.split(passage_separator, all_text.strip())
        # Since this is the end of the method, the logic may as well return the list comprehension result
        # rather than spend the extra effort to modify the existing passage list and then return the result.
        if self.strip_excess_whitespace_from_list:
            return [passage.strip() for passage in passage_list]
        return passage_list



# This file is used as a way of defining accessible modules when importing the utilities library.
# Note that common is not included in the default module import, since it's mostly intended for internal use.

from meaningless.utilities import yaml_file_interface, json_file_interface, xml_file_interface, csv_file_interface, \
    exceptions



# These are custom exceptions used for particular failure cases


class BaseError(Exception):
    pass


class UnsupportedTranslationError(BaseError):
    """
    An exception thrown when handling translations that are not currently supported
    """

    def __init__(self, translation):
        """
        :param translation: Translation code. For example, 'NIV', 'ESV', 'NLT'
        :type translation: str
        """
        super(UnsupportedTranslationError, self).__init__(f'{translation} is an unsupported translation')
        self.translation = translation


class InvalidPassageError(BaseError):
    """
    An exception thrown when processing a non-existent passage (or passage range)
    """

    def __init__(self, book, chapter_from, passage_from, chapter_to, passage_to, translation):
        """
        :param book: Name of the book
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param passage_from: First passage number to get in the first chapter
        :type passage_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param passage_to: Last passage number to get in the last chapter
        :type passage_to: int
        :param translation: Translation code. For example, 'NIV', 'ESV', 'NLT'
        :type translation: str
        """
        if chapter_from == chapter_to:
            if passage_from == passage_to:
                passage = f'{chapter_from}:{passage_from}'
            else:
                passage = f'{chapter_from}:{passage_from} - {passage_to}'
        else:
            passage = f'{chapter_from}:{passage_from} - {chapter_to}:{passage_to}'
        super(InvalidPassageError, self).__init__(
            f'{book} {passage} is an invalid passage in the {translation} translation')
        self.book = book
        self.chapter_from = chapter_from
        self.passage_from = passage_from
        self.chapter_to = chapter_to
        self.passage_to = passage_to
        self.translation = translation


class InvalidSearchError(BaseError):
    """
    An exception thrown when searching for an invalid passage on the Bible Gateway site
    """

    def __init__(self, url):
        """
        :param url: The URL which contains the invalid search results
        :type url: str
        """
        super(InvalidSearchError, self).__init__(f'Failed to retrieve search results from {url}')
        self.url = url


class TranslationMismatchError(BaseError):
    """
    An exception thrown when using the extractor to read a file, and both use different translations
    """

    def __init__(self, extractor_translation, file_translation):
        """
        :param extractor_translation: Translation code used by the extractor. For example, 'NIV', 'ESV', 'NLT'
        :type extractor_translation: str
        :param file_translation: Translation code used in the file. For example, 'NIV', 'ESV', 'NLT'
        :type file_translation: str
        """
        super(TranslationMismatchError, self).__init__(
            f'The extractor is using the {extractor_translation} translation, '
            f'but attempted to read a file in the {file_translation} translation')
        self.extractor_translation = extractor_translation
        self.file_translation = file_translation



import os
from meaningless.bible_base_downloader import BaseDownloader
from meaningless.utilities import json_file_interface


class JSONDownloader(BaseDownloader):
    """
    An downloader object that stores Bible passages into a local JSON file
    """

    def __init__(self, translation='NIV', show_passage_numbers=True, default_directory=os.getcwd(),
                 strip_excess_whitespace=False, enable_multiprocessing=True, use_ascii_punctuation=False):
        super().__init__(json_file_interface.write, translation, show_passage_numbers, default_directory,
                         strip_excess_whitespace, enable_multiprocessing, use_ascii_punctuation,
                         file_extension='.json', write_key_as_string=False)



import os
import json

# This is a collection of common methods used for interacting with JSON files.


def write(data_file, document):
    """
    A helper function to write to a JSON data file.

    :param data_file: Path to the data file to write to
    :type data_file: str
    :param document: In-memory JSON structure, usually a dictionary
    :type document: dict
    :return: Returns 1 on success. Raises an exception when a write problem occurs.
    :rtype: int
    """
    # Only create the directory if it doesn't already exist. This is also to account for directories which are the
    # top level of a given drive (e.g. C:/) which can't be created by the file system due to denied access.
    data_directory = os.path.dirname(data_file)
    if not os.path.exists(data_directory):
        os.makedirs(data_directory, exist_ok=True)
    # Use UTF-8 encoding to allow for Unicode characters to be written to the file
    with open(data_file, 'w', newline='', encoding='utf-8') as file:
        json.dump(document, file, sort_keys=True, indent=2, ensure_ascii=False)
    return 1


def read(data_file):
    """
    A helper function to read a JSON data file

    :param data_file: Path the data file to read
    :type data_file: str
    :return: Contents of the file as an object. Raises an exception when a read problem occurs.
    :rtype: dict
    """
    # Use UTF-8 encoding to be able to read Unicode characters
    with open(data_file, 'r', encoding='utf-8') as file:
        contents = json.load(file)
    return contents



import re
import os
import xmltodict

# This is a collection of common methods used for interacting with XML files.


def __get_root_name():
    """
    Gets the top-level tag name used in the document.

    :return: Returns the tag name.
    :rtype: str
    """
    return 'root'


def __get_book_name():
    """
    Gets the tag name used for books in the document.

    :return: Returns the tag name.
    :rtype: str
    """
    return 'book'


def __get_chapter_name():
    """
    Gets the tag name used for chapters in the document.

    :return: Returns the tag name.
    :rtype: str
    """
    return 'chapter'


def __get_passage_name():
    """
    Gets the tag name used for passages in the document.

    :return: Returns the tag name.
    :rtype: str
    """
    return 'passage'


def __get_numeric_prefix():
    """
    Gets the prefix added onto tag names with only numeric characters to make them valid XML tags.

    :return: Returns the tag prefix.
    :rtype: str
    """
    return '_'


def __get_space_placeholder():
    """
    Gets the placeholder character for representing spaces in XML tags.

    :return: Returns the placeholder string
    :rtype: str
    """
    return __get_numeric_prefix()


def __get_cdata_key():
    """
    Gets the placeholder string for the cdata key that is automatically added into a parsed XML document

    :return: Returns the key name
    :rtype: str
    """
    return ''


def __restore_xml_key_and_value(path, key, value):
    """
    A post-processing function when parsing the XML document to normalise keys into a more familiar format.

    :param path: Path from the document root to the key-value pair. Required as part of the xmltodict interface.
    :type path: str
    :param key: Dictionary key in the document. Required as part of the xmltodict interface.
    :type key: str
    :param value: Dictionary value in the document. Required as part of the xmltodict interface.
    :type value: str
    :return: Returns the modified key-value pair
    :rtype: tuple
    """
    # Revert the space placeholder with an actual space, since the limitations of XML tag naming are removed.
    # Since the space placeholder is a reserved string, any legitimate use for this in a tag name is ignored.
    # Strip excess whitespace afterwards to remove converted space placeholders at the start or end of a string.
    new_key = key.replace(__get_space_placeholder(), ' ').replace(__get_numeric_prefix(), '').strip()
    # Most keys can be reverted back to their original case, as XML conventions no longer apply
    if new_key not in [__get_root_name()]:
        new_key = new_key.title()
    # Omit keys that are empty or consisted only of space placeholders, which would have reduced to an empty string.
    # This is also a convenient way to remove the generated cdata key, by setting it to something that can be ignored.
    if len(new_key) <= 0:
        return
    # Reading a tag without contents resolve to None, which is incorrect in the case of omitted passages
    new_value = value if value is not None else ''
    return new_key, new_value


def write(data_file, document):
    """
    A helper function to write to a XML data file.
    Note that the input data must adhere to the following conventions:

    1. If a top-level key is named 'Info', it must be a mapping of string keys to string values.
    2. All other top-level keys map to a 3-layer dictionary or similar data structure with key-value pairs.
    3. The 3rd layer of this data structure is a mapping of string keys to string values.

    :param data_file: Path to the data file to write to
    :type data_file: str
    :param document: In-memory data structure, usually a dictionary
    :type document: dict
    :return: Returns 1 on success. Raises an exception when a write problem occurs.
    :rtype: int
    """
    # Only create the directory if it doesn't already exist. This is also to account for directories which are the
    # top level of a given drive (e.g. C:/) which can't be created by the file system due to denied access.
    data_directory = os.path.dirname(data_file)
    if not os.path.exists(data_directory):
        os.makedirs(data_directory, exist_ok=True)

    modified_document = {__get_root_name(): {}}
    for key in list(document.keys()):
        # Copy in all the "non-book" information as is
        if key in ['Info']:
            # Convert all immediate keys to lowercase, which is the general XML convention
            new_key = key.lower()
            modified_document[__get_root_name()][new_key] = {}
            for sub_key in document[key]:
                modified_document[__get_root_name()][new_key][sub_key.lower()] = document[key][sub_key]
            continue

        # The XML output should provide some supplementary metadata for use in the read() function
        # Note that the leading prefix is only required for book names which start with a number, but add it in
        # all cases anyway for simpler logic.
        modified_document[__get_root_name()][__get_book_name()] = {
            '@name': key,
            '@tag': f'{__get_space_placeholder()}{key.replace(" ", __get_space_placeholder())}',
            __get_chapter_name(): []
        }
        for chapter in list(document[key].keys()):
            modified_document[__get_root_name()][__get_book_name()][__get_chapter_name()].append({
                '@number': chapter,
                '@tag': f'{__get_numeric_prefix()}{chapter}',
                __get_passage_name(): []
            })
            for passage in list(document[key][chapter].keys()):
                # Avoid potential unparsing errors by always casting the passage content as a string
                modified_document[__get_root_name()][__get_book_name()][__get_chapter_name()][-1][__get_passage_name()]\
                    .append({
                        '@number': passage,
                        '@tag': f'{__get_numeric_prefix()}{passage}',
                        '#text': str(document[key][chapter][passage])
                    })
    # Use space indentation, to keep in line with most of the other file interfaces
    contents = xmltodict.unparse(modified_document, pretty=True, indent='  ')
    # Use UTF-8 encoding to allow for Unicode characters to be written to the file
    with open(data_file, 'w', newline='', encoding='utf-8') as file:
        file.write(contents)
    return 1


def read(data_file):
    """
    A helper function to read a XML data file.
    Note that the input data must adhere to the following conventions:

    1. The top-level tag is called 'root'.
    2. All <book> tags must have a "tag" attribute.
    3. All <chapter> tags must have a "tag" attribute and must be nested within a <book> tag.
    4. All <passage> tags must have a "tag" attribute and must be nested within a <chapter> tag.

    If the XML data file is valid, the returned object will differ from the XML data file in the following ways:

    1. Leading and trailing underscores will be removed.
    2. All other underscores will be converted into spaces.
    3. All keys will be converted to title case (all first letters of each word are capitalised)

    :param data_file: Path to the data file to read
    :type data_file: str
    :return: Contents of the file as an object. Raises an exception when a read problem occurs.
    :rtype: dict
    """
    # Use UTF-8 encoding to be able to read Unicode characters
    with open(data_file, 'r', encoding='utf-8') as file:
        contents = file.read()
    # Pre-process the XML contents such that when it gets parsed, it is mostly in the expected structure
    # that file interfaces are to operate by.
    #
    # This works by pre-computing the relevant object keys as XML-compliant values when writing
    # the file and restoring that value as the preferred tag name when reading it back.
    # After parsing, it's necessary to do some post-processing on the document again, as most of the
    # XML-compliant modifications are only needed to pass the underlying XML validation checks.
    tags_with_predefined_replacements = [__get_book_name(), __get_chapter_name(), __get_passage_name()]
    for tag in tags_with_predefined_replacements:
        contents = re.sub(fr'<{tag}\s*.*?\s*tag="(.+?)">(.*?)</{tag}>', r'<\1>\2</\1>',
                          contents, flags=re.DOTALL)
    # Since white space and newlines are preserved, the side effect is that a cdata key is automatically added into
    # each level of the document, with its value being spaces and newline characters.
    raw_document = xmltodict.parse(contents, strip_whitespace=False, cdata_key=__get_cdata_key(),
                                   postprocessor=__restore_xml_key_and_value)
    return raw_document[__get_root_name()]



import os
import re
from meaningless.utilities import common
from meaningless.utilities.exceptions import UnsupportedTranslationError, InvalidPassageError, TranslationMismatchError


class BaseExtractor:
    """
    An base extractor object that retrieves Bible passages from a file
    """

    def __init__(self, file_reading_function, translation='NIV', show_passage_numbers=True, output_as_list=False,
                 strip_excess_whitespace_from_list=False, default_directory=os.getcwd(),
                 use_ascii_punctuation=False, file_extension='', read_key_as_string=False):
        """
        :param file_reading_function: Function definition used to specify how to read a given file.
                                      The function should only take 1 argument, which states the file path to read.
        :type file_reading_function: callable[[str], dict]
        :param translation: Translation code for the particular passage. For example, 'NIV', 'ESV', 'NLT'
        :type translation: str
        :param show_passage_numbers: If True, any present passage numbers are preserved. Defaults to True.
        :type show_passage_numbers: bool
        :param output_as_list: When True, returns the passage data as a list of strings. Defaults to False.
        :type output_as_list: bool
        :param strip_excess_whitespace_from_list: When True and output_as_list is also True, leading and trailing
                                                  whitespace characters are removed for each string element in the list.
                                                  Defaults to False.
        :type strip_excess_whitespace_from_list: bool
        :param default_directory: Directory containing the file to read from.
                                  Defaults to the current working directory.
        :type default_directory: str
        :param use_ascii_punctuation: When True, converts all Unicode punctuation characters into their ASCII
                                      counterparts. This also applies to passage separators. Defaults to False.
        :type use_ascii_punctuation: bool
        :param file_extension: File extension used when reading from a default file when file_path is not provided.
        :type file_extension: str
        :param read_key_as_string: If True, specifies that all keys in the extracted file will be strings.
               Defaults to False.
        :type read_key_as_string: bool
        """
        self.translation = translation
        self.show_passage_numbers = show_passage_numbers
        self.output_as_list = output_as_list
        self.strip_excess_whitespace_from_list = strip_excess_whitespace_from_list
        self.default_directory = default_directory
        self.use_ascii_punctuation = use_ascii_punctuation
        self.file_extension = file_extension
        self.file_reading_function = file_reading_function
        self.read_key_as_string = read_key_as_string

    def get_passage(self, book, chapter, passage, file_path=''):
        """
        Gets a single passage from a file.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage: Passage number
        :type passage: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: The specified passage. Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter, passage, chapter, passage, file_path)

    def get_passages(self, book, chapter, passage_from, passage_to, file_path=''):
        """
        Gets a range of passages of the same chapter from a file.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage_from: First passage number to get
        :type passage_from: int
        :param passage_to: Last passage number to get
        :type passage_to: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: The passages between the specified passages (inclusive). Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter, passage_from, chapter, passage_to, file_path)

    def get_chapter(self, book, chapter, file_path=''):
        """
        Gets a single chapter from a file.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: All passages in the chapter. Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter, 1, chapter, common.get_end_of_chapter(), file_path)

    def get_chapters(self, book, chapter_from, chapter_to, file_path=''):
        """
        Gets a range of passages from a specified chapter selection from a file.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param book: Name of the book
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: All passages between the specified chapters (inclusive). Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_passage_range(book, chapter_from, 1, chapter_to, common.get_end_of_chapter(), file_path)

    def get_book(self, book, file_path=''):
        """
        Gets all chapters for a specific book from a file.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param book: Name of the book
        :type book: str
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: All passages in the specified book. Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        return self.get_chapters(book, 1, common.get_chapter_count(book, self.translation), file_path)

    def get_passage_range(self, book, chapter_from, passage_from, chapter_to, passage_to, file_path=''):
        """
        Gets a range of passages from one specific passage to another passage from a file.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param book: Name of the book
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param passage_from: First passage number to get in the first chapter
        :type passage_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param passage_to: Last passage number to get in the last chapter
        :type passage_to: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :return: All passages between the specified passages (inclusive). Empty string/list if the passage is invalid.
        :rtype: str or list
        """
        translation = self.translation.upper()
        if common.is_unsupported_translation(translation):
            raise UnsupportedTranslationError(translation)
        # Standardise letter casing to ensure key access errors are not caused by case sensitivity
        book_name = book.title()
        if len(file_path) > 0:
            file_to_read = file_path
        else:
            file_to_read = os.path.join(self.default_directory, f'{book_name}{self.file_extension}')
        document = self.file_reading_function(file_to_read)
        passage_list = []
        # Fail-fast on invalid passages
        if not document:
            raise InvalidPassageError(book_name, chapter_from, passage_from, chapter_to, passage_to, translation)
        # Ensure matching translations are used in the extractor and file to ensure the correct translation is used.
        # Otherwise, the extractor could be used to read a bunch of files and each return different results
        # on the same passage even though the translation class property is unchanged.
        file_translation = document['Info']['Translation']
        if not common.is_matching_translation(translation, file_translation):
            raise TranslationMismatchError(translation, file_translation)

        # Apply a boundary to the chapters to prevent invalid keys being accessed
        # Use the last key of the book, as it's not guaranteed that the number of chapters == last key
        chapter_first = common.dict_keys_to_sorted_list(document[book_name].keys())[0]
        chapter_final = common.dict_keys_to_sorted_list(document[book_name].keys())[-1]
        capped_chapter_from = common.get_capped_integer(chapter_from, min_value=chapter_first, max_value=chapter_final)
        capped_chapter_to = common.get_capped_integer(chapter_to, min_value=chapter_first, max_value=chapter_final)
        # Extend the range by 1 since chapter_to is also included in the iteration
        for chapter in range(capped_chapter_from, capped_chapter_to + 1):
            # Determine the range of passages to extract from the chapter
            passage_initial = 1
            # Use the last key of the chapter section, as it's not guaranteed that the number of passages == last key
            # Cast to an integer, as it is used in certain numeric operations later on.
            passage_final = int(common.dict_keys_to_sorted_list(document[book_name][self.__key_cast(chapter)].keys())
                                [-1])
            passage_min = int(common.dict_keys_to_sorted_list(document[book_name][self.__key_cast(chapter)].keys())[0])
            if chapter == capped_chapter_from:
                # For the first chapter, an initial set of passages can be ignored (undercuts the passage selection)
                # Apply a boundary to the passage to prevent invalid keys being accessed
                passage_initial = common.get_capped_integer(passage_from, min_value=passage_min,
                                                            max_value=passage_final)
            if chapter == capped_chapter_to:
                # For the last chapter, a trailing set of passages can be ignored (exceeds the passage selection)
                # Apply a boundary to the passage to prevent invalid keys being accessed
                passage_final = common.get_capped_integer(passage_to, min_value=passage_min, max_value=passage_final)
            # Extend the range by 1 since the last passage is also included in the iteration
            [passage_list.append(document[book_name][self.__key_cast(chapter)][self.__key_cast(passage)]) for passage in
             range(passage_initial, passage_final + 1)]
            # Start each chapter on a new line when outputting as a string.
            if not self.output_as_list:
                passage_list.append('\n')

        if self.use_ascii_punctuation:
            passage_list = [common.unicode_to_ascii_punctuation(passage) for passage in passage_list]
        if not self.show_passage_numbers:
            # This assumes that all superscript numbers are indicative of the passage number
            passage_list = [common.remove_superscript_numbers_in_passage(passage) for passage in passage_list]
        if self.output_as_list:
            if self.strip_excess_whitespace_from_list:
                return [passage.strip() for passage in passage_list]
            return passage_list
        # Convert the list of passages into a string, as strings are immutable and manually re-initialising a new string
        # in the loop can be costly to performance.
        all_text = ''.join([passage for passage in passage_list])

        return all_text.strip()

    def find_text_in_passage(self, search_text, book, chapter, passage, file_path='',
                             is_case_sensitive=False, is_regex=False):
        """
        Gets a single passage from a file, containing a certain string or matching a particular regular expression.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param search_text: Search text
        :type search_text: str
        :type book: str
        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage: Passage number
        :type passage: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :param is_case_sensitive: When True, the passage will attempt to match against the search text or regular
                                  expression with case sensitivity. Defaults to False.
        :type is_case_sensitive: bool
        :param is_regex: When True, the search text will be interpreted as a regular expression. Defaults to False.
        :type is_regex: bool
        :return: The specified passage. Empty string/list if the passage is invalid or the passage did not meet
                 the search criteria.
        :rtype: str or list
        """
        return self.find_text_in_passage_range(search_text, book, chapter, passage, chapter,
                                               passage, file_path, is_case_sensitive, is_regex)

    def find_text_in_passages(self, search_text, book, chapter, passage_from, passage_to, file_path='',
                              is_case_sensitive=False, is_regex=False):
        """
        Gets a range of passages of the same chapter from a file, where each passage contains
        a certain string or matches a particular regular expression.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param search_text: Search text
        :type search_text: str
        :type book: str
        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param passage_from: First passage number to get
        :type passage_from: int
        :param passage_to: Last passage number to get
        :type passage_to: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :param is_case_sensitive: When True, passages will attempt to match against the search text or regular
                                  expression with case sensitivity. Defaults to False.
        :type is_case_sensitive: bool
        :param is_regex: When True, the search text will be interpreted as a regular expression. Defaults to False.
        :type is_regex: bool
        :return: All passages between the specified passages (inclusive). Empty string/list if the passage is invalid
                 or none of the passages met the search criteria.
        :rtype: str or list
        """
        return self.find_text_in_passage_range(search_text, book, chapter, passage_from, chapter,
                                               passage_to, file_path, is_case_sensitive, is_regex)

    def find_text_in_chapter(self, search_text, book, chapter, file_path='', is_case_sensitive=False, is_regex=False):
        """
        Gets a range of passages from one specific passage to another passage from a file, where each passage contains
        a certain string or matches a particular regular expression.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param search_text: Search text
        :type search_text: str
        :type book: str
        :param book: Name of the book
        :type book: str
        :param chapter: Chapter number
        :type chapter: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :param is_case_sensitive: When True, passages will attempt to match against the search text or regular
                                  expression with case sensitivity. Defaults to False.
        :type is_case_sensitive: bool
        :param is_regex: When True, the search text will be interpreted as a regular expression. Defaults to False.
        :type is_regex: bool
        :return: All passages between the specified passages (inclusive). Empty string/list if the passage is invalid
                 or none of the passages met the search criteria.
        :rtype: str or list
        """
        return self.find_text_in_passage_range(search_text, book, chapter, 1, chapter,
                                               common.get_end_of_chapter(), file_path, is_case_sensitive, is_regex)

    def find_text_in_chapters(self, search_text, book, chapter_from, chapter_to, file_path='',
                              is_case_sensitive=False, is_regex=False):
        """
        Gets a range of passages from a specified chapter selection from a file, where each passage contains
        a certain string or matches a particular regular expression.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param search_text: Search text
        :type search_text: str
        :type book: str
        :param book: Name of the book
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :param is_case_sensitive: When True, passages will attempt to match against the search text or regular
                                  expression with case sensitivity. Defaults to False.
        :type is_case_sensitive: bool
        :param is_regex: When True, the search text will be interpreted as a regular expression. Defaults to False.
        :type is_regex: bool
        :return: All passages between the specified passages (inclusive). Empty string/list if the passage is invalid
                 or none of the passages met the search criteria.
        :rtype: str or list
        """
        return self.find_text_in_passage_range(search_text, book, chapter_from, 1, chapter_to,
                                               common.get_end_of_chapter(), file_path, is_case_sensitive, is_regex)

    def find_text_in_book(self, search_text, book, file_path='', is_case_sensitive=False, is_regex=False):
        """
        Gets all passages for a specific book from a file, where each passage contains
        a certain string or matches a particular regular expression.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param search_text: Search text
        :type search_text: str
        :param book: Name of the book
        :type book: str
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :param is_case_sensitive: When True, passages will attempt to match against the search text or regular
                                  expression with case sensitivity. Defaults to False.
        :type is_case_sensitive: bool
        :param is_regex: When True, the search text will be interpreted as a regular expression. Defaults to False.
        :type is_regex: bool
        :return: All passages between the specified passages (inclusive). Empty string/list if the passage is invalid
                 or none of the passages met the search criteria.
        :rtype: str or list
        """
        return self.find_text_in_passage_range(search_text, book, 1, 1, common.get_chapter_count(book, self.translation),
                                               common.get_end_of_chapter(), file_path, is_case_sensitive, is_regex)

    def find_text_in_passage_range(self, search_text, book, chapter_from, passage_from, chapter_to, passage_to,
                                   file_path='', is_case_sensitive=False, is_regex=False):
        """
        Gets a range of passages from one specific passage to another passage from a file, where each passage contains
        a certain string or matches a particular regular expression.

        Output will be automatically adjusted to the chapter and passage boundaries from the input file.

        :param search_text: Search text
        :type search_text: str
        :type book: str
        :param book: Name of the book
        :type book: str
        :param chapter_from: First chapter number to get
        :type chapter_from: int
        :param passage_from: First passage number to get in the first chapter
        :type passage_from: int
        :param chapter_to: Last chapter number to get
        :type chapter_to: int
        :param passage_to: Last passage number to get in the last chapter
        :type passage_to: int
        :param file_path: When specified, reads the file from this location with a custom filename and extension.
                          Using this parameter will take priority over the default_directory class property.
                          Defaults to the default_directory path with the book as the file name with a default
                          extension.
        :type file_path: str
        :param is_case_sensitive: When True, passages will attempt to match against the search text or regular
                                  expression with case sensitivity. Defaults to False.
        :type is_case_sensitive: bool
        :param is_regex: When True, the search text will be interpreted as a regular expression. Defaults to False.
        :type is_regex: bool
        :return: All passages between the specified passages (inclusive). Empty string/list if the passage is invalid
                 or none of the passages met the search criteria.
        :rtype: str or list
        """
        # Temporarily force output as a list to process individual passages more easily
        # Note that an exception occurring during passage retrieval will result in self.output_as_list not reverting
        # to its original value. Any logic that catches raised exceptions for this function may be required to manually
        # restore the original value for self.output_as_list, depending on the exception handling logic being used.
        output_as_list = self.output_as_list
        self.output_as_list = True
        passages = self.get_passage_range(book, chapter_from, passage_from, chapter_to, passage_to, file_path)
        self.output_as_list = output_as_list

        matching_passages = []
        if is_case_sensitive:
            keyword_regex = re.compile(search_text)
        else:
            keyword_regex = re.compile(search_text, re.IGNORECASE)
        for passage in passages:
            if (not is_case_sensitive and search_text.casefold() in passage.casefold()) or \
               (is_case_sensitive and search_text in passage) or \
               (is_regex and re.search(keyword_regex, passage)):
                if not self.output_as_list:
                    # Strip whitespace on string output to match the same kind of returned data as get_passage_range
                    matching_passages.append(passage.strip())
                else:
                    matching_passages.append(passage)

        if self.output_as_list:
            return matching_passages
        return '\n'.join(matching_passages)

    def __key_cast(self, key):
        """
        A helper function to cast a dictionary key to a string or an integer.

        :param key: Dictionary key
        :type key: int or str
        :return: Type-casted dictionary key
        :rtype: int or str
        """
        return common.cast_to_str_or_int(key, self.read_key_as_string)
